/* 
    added by Joerg Schulenburg 2006
    to switch between super computers (with slow disk and enough memory) 
    and small PCs (where we dont have enough memory) define USE_FILE
    to store and read the terms sequentially
 */

#include <stdlib.h>
#include <stdio.h>
#include <string.h>
#include "Externals.h"
extern int myid;

#ifdef USE_FILE
FILE *fp[2]={NULL,NULL};           /* 0=gs_terms, 1=terms */
#define  BUFSIZE (1024*1024l*8)
char buf[2][BUFSIZE];           // file cache
#else
// its a good idea to set ADDLEN to Maxmemory/32 to avoid framentation+copy
// ToDo: make a list of storage blocks to avoid realloc (fragmentation+copy)
#define ADDLEN (1024*1024l)     // num of additional terms for realloc
#define DESCSIZE (sizeof(int)*6)      // descriptors size
#define MPLYSIZE (sizeof(double))     // multiplier size
#define TERMSIZE (DESCSIZE+MPLYSIZE)  // one term size
char   *t_buf[2]={NULL, NULL};              // storage buffer
size_t  t_len[2]={0,0}, // allocated memory
        t_end[2]={0,0}, // index to next free element
        t_idx[2]={0,0}; // index to next element
#endif

// sizeof(size_t):      4 (=2e9B)         8 (=9e18B)
//    i386              cc        
//    sparc             cc,gcc        (cc -xarch=v9)
//    alphaev6                         gcc       
//    mips              c89,gcc       (c89 -64), (gcc -mabi)      

           /* id: 0=gs_terms, 1=terms */
#ifdef USE_FILE
int term_open(int id, char *opt){
  char name[256];
  strcpy(name,Control.name); strcat(name,((id==0)?"_gs":".terms"));
  if (myid==0) printf(" fopen(%s) - slow USE_FILE mode!\n",name);
  fp[id]=fopen(name,opt);
  if (!fp[id]) {
    printf("Cannot Open %s for %s...\nExiting!!\n\n", name, opt);
    exit(1);
  }
  setvbuf(fp[id],buf[id],_IOFBF,BUFSIZE);
  return 0;
}
#endif

void term_rewind(int id) {
#ifdef USE_FILE
  if (!fp[id]) { term_open(id,"r"); }
  rewind(fp[id]);
#else
  t_idx[id]=0;
#endif
}

/* set error if storing was not possible,
   skip further storing, break getting  */
int error=0;

           /* id: 0=gs_terms, 1=terms */
int term_store(int id, int *descriptors, double multiplier) {
#ifdef USE_FILE
  int nio;
  if (error) return 0; // skip on error
  if (!fp[id]) { term_open(id,"w+"); }
  nio =fwrite(descriptors,sizeof(   int),6,fp[id]);
  nio+=fwrite(&multiplier,sizeof(double),1,fp[id]);
  if (nio!=7) {
    error=1;
    printf("%3d: term_store fwrite_error=%d L%d\n",
                 myid, ferror(fp[id]), __LINE__);
    printf(" Terms can not be stored, you will get no results!\n"
           " Program will stop after all terms are counted.\n");
  }
#else
  if (error) return 0; // skip on error
  if   (t_end[id] + TERMSIZE       >= t_len[id]) {
    if (t_len[id] + TERMSIZE*ADDLEN < t_len[id]) {
         printf("Error: term_store, size_t overflow\nexit\n"); exit(1); }
    if (((int)t_len[id]) + TERMSIZE*ADDLEN < (int)t_len[id]) {
         printf("Warning: term_store, int overflow\n"); }
    t_len[id] += ADDLEN*TERMSIZE;        // extend the memory buffer
    // realloc has the problem of fragmentation and copying (max RAM/2)
    t_buf[id]=(char *)realloc(t_buf[id], t_len[id]);
    if (!t_buf[id]) {
      printf("%3d: realloc failed, len= %10.2e [+=%.2e*%d]\n",
             myid, 1.*t_len[id], 1.*ADDLEN, TERMSIZE);
      printf(" Further terms can not be stored, you will get no results!\n"
             " Program will stop after all terms are counted.\n");
      printf(" If you have fast storage, try USE_FILE which is slower\n");
      error=2;
      return -1;
    }
  }
  memcpy(t_buf[id]+t_end[id], descriptors, DESCSIZE); t_end[id]+=DESCSIZE;
  memcpy(t_buf[id]+t_end[id], &multiplier, MPLYSIZE); t_end[id]+=MPLYSIZE;
#endif
  return 0;
}

           /* id: 0=gs_terms, 1=terms */
// returning 0 on OK, else exit
int term_get(int id, int *descriptors, double *multiplier) {
  if (error) {
     printf("Error term_get: Not all terms could be stored!\n"
            "Error term_get: Take the counters to estimate what you need\n");
     exit(1);
     // ToDo: repeat slow calculation of a single element here?
  }
#ifdef USE_FILE
  int nio;
  if (!fp[id]) { term_open(id,"r"); }
  nio =fread(descriptors,sizeof(   int),6,fp[id]);
  nio+=fread(multiplier ,sizeof(double),1,fp[id]);
  if (nio!=7) {
    printf("%3d: term_get fread_error=%d L%d\n",
                 myid, ferror(fp[id]), __LINE__); exit(1); }
#else
  if (t_idx[id] + TERMSIZE > t_end[id]) {
    printf("%3d: access beyond t_buf\n",myid); exit(1); }
  memcpy(descriptors,t_buf[id]+t_idx[id], DESCSIZE); t_idx[id]+=DESCSIZE;
  memcpy(multiplier ,t_buf[id]+t_idx[id], MPLYSIZE); t_idx[id]+=MPLYSIZE;
#endif
  return 0;
}

// Test:
//  g++ -g -DTEST -Wall storage.C
//  valgrind -v --tool=memcheck ./a.out
#ifdef TEST
int myid=0;
int main() {
  int i;
  int x[6]={0,1,2,3,4,5};
  double y=1.2345;
  for (i=0;i<10;i++) term_store(0,x,y);
  for (i=0;i<11;i++) term_get(0,x,&y);
}
#endif
